import tempfile
from enum import Enum
from tcheckerpy.utils import call_tchecker

class Relationship(Enum):
    STRONG_TIMED_BISIM = 0 # strong timed bisimilarity

def compare(first_sys_decl: str, second_sys_decl: str, relationship: Relationship = Relationship.STRONG_TIMED_BISIM, 
            generate_witness = False, block_size: int | None = None, table_size: int | None = None) -> tuple[bool, str, str]:
    """
    Checks for bisimilarity of two timed automata.

    :param first_sys_decl: system declaration of first timed automaton
    :param second_sys_decl: system declaration of second timed automaton
    :param relationship: relationship to be checked (so far only strong timed bisimilarity is supported)
    :param block_size: block size for internal computation
    :param table_size: table size for internal computation
    :return: result of bisimilarity check, statistics and witness
    """
    
    with tempfile.NamedTemporaryFile(delete=False) as temp_file_first_sysdecl:
        temp_file_first_sysdecl.write(first_sys_decl.encode('utf-8'))
        temp_file_path_first_sysdecl = temp_file_first_sysdecl.name
    with tempfile.NamedTemporaryFile(delete=False) as temp_file_second_sysdecl:
        temp_file_second_sysdecl.write(second_sys_decl.encode('utf-8'))
        temp_file_path_second_sysdecl  = temp_file_second_sysdecl.name
        
    # call TChecker function
    stats, witness = call_tchecker.call_tchecker_function_in_new_process(
        func_name="tck_compare",
        argtypes=["ctypes.c_char_p", "ctypes.c_char_p", "ctypes.c_int",
                  "ctypes.POINTER(ctypes.c_int)", "ctypes.POINTER(ctypes.c_int)", "ctypes.c_bool"],
        has_result=True,
        args=[temp_file_path_first_sysdecl, temp_file_path_second_sysdecl, relationship.value, 
              block_size, table_size, generate_witness]
    )

    return "RELATIONSHIP_FULFILLED true" in stats, stats, witness



import tempfile
from enum import Enum
from tcheckerpy.utils import call_tchecker

class __SimulationType(Enum):
    INTERACTIVE = 0 # interactive simulation
    ONE_STEP = 1 # one-step simulation (output initial or next states)
    RANDOMIZED = 2 # randomized simulation

def __call_tck_simulate(sys_decl: str, simulation_type: __SimulationType, nsteps: int,
                        starting_state: str | None = None) -> str:
    
    if simulation_type == __SimulationType.RANDOMIZED and nsteps == None:
        raise ValueError("Randomized simulation requires number of steps")
    
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(sys_decl.encode('utf-8'))
        temp_file_path = temp_file.name

    # call TChecker function
    _, result = call_tchecker.call_tchecker_function_in_new_process(
        func_name="tck_simulate",
        argtypes=["ctypes.c_char_p", "ctypes.c_int", "ctypes.c_int", "ctypes.c_char_p", 
                  "ctypes.POINTER(ctypes.c_int)", "ctypes.c_bool"],
        has_result=True,
        args=[temp_file_path, simulation_type.value, 1, starting_state or "",
              nsteps, not simulation_type == __SimulationType.ONE_STEP] 
    )

    return result

def one_step_simulation(sys_decl: str, starting_state: str | None = None) -> str:
    """
    Simulates one step of timed automaton, starting from the given state.

    :param sys_decl: system declaration of timed automaton
    :param starting_state: starting state, specified as a JSON object with keys vloc, intval and zone
                           vloc: comma-separated list of location names (one per process), in-between < and >
                           intval: comma-separated list of assignments (one per integer variable)
                           zone: conjunction of clock-constraints (following TChecker expression syntax)
    :return: initial states if `starting_state` is none, next states otherwise
    """
    return __call_tck_simulate(sys_decl, __SimulationType.ONE_STEP, 0, starting_state)

def randomized_simulation(sys_decl: str, nsteps: int, starting_state: str | None = None) -> str:
    """
    Randomly simulates steps of timed automaton, starting from the given state.

    :param sys_decl: system declaration of timed automaton
    :param starting_state: starting state, specified as a JSON object with keys vloc, intval and zone
                           vloc: comma-separated list of location names (one per process), in-between < and >
                           intval: comma-separated list of assignments (one per integer variable)
                           zone: conjunction of clock-constraints (following TChecker expression syntax)
    :param nsteps: number of steps to simulate
    :return: simulation trace
    """
    return __call_tck_simulate(sys_decl, __SimulationType.RANDOMIZED, nsteps, starting_state)


import tempfile
from enum import Enum
from tcheckerpy.utils import call_tchecker

class Algorithm(Enum):
    REACH = 0 # standard reachability algorithm over the zone graph
    CONCUR19 = 1 # reachability algorithm over the local-time zone graph, with sync-subsumption
    COVREACH = 2 # reachability algorithm over the zone graph with inclusion subsumption
    ALU_COVREACH = 3 # reachability algorithm over the zone graph with aLU subsumption

class SearchOrder(Enum):
    BFS = "bfs" # breadth-first search
    DFS = "dfs" # depth-first search

class Certificate(Enum):
    GRAPH = 0 # graph of explored state-space
    SYMBOLIC = 1 # symbolic run to a state with searched labels if any
    CONCRETE = 2 # concrete run to a state with searched labels if any (only for reach and covreach)
    NONE = 3 # no certificate

def reach(sys_decl: str, algorithm: Algorithm, search_order: SearchOrder = SearchOrder.BFS, 
          certificate: Certificate = Certificate.NONE, labels: list[str] = [],
          block_size: int | None = None, table_size: int | None = None) -> tuple[bool, str, str]:
    """
    Checks for reachability of timed automaton.

    :param sys_decl: system declaration of timed automaton
    :param algorithm: algorithm to be used (see `tck_reach.Algorithm`)
    :param search_order: search order to be used (see `tck_reach.SearchOrder`)
    :param certificate: certificate type (see `tck_reach.Certificate`)
    :param labels: list of accepting labels
    :param block_size: block size for internal computation
    :param table_size: table size for internal computation
    :return: result of reachability check, statistics and certificate
    """
    
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(sys_decl.encode('utf-8'))
        temp_file_path = temp_file.name

    # convert list to string
    labels_str = ", ".join(labels)
        
    # call TChecker function
    stats, cert = call_tchecker.call_tchecker_function_in_new_process(
        func_name="tck_reach",
        argtypes=["ctypes.c_char_p", "ctypes.c_char_p", "ctypes.c_int", "ctypes.c_char_p",
                  "ctypes.c_int", "ctypes.POINTER(ctypes.c_int)", "ctypes.POINTER(ctypes.c_int)"],
        has_result=True,
        args=[temp_file_path, labels_str, algorithm.value, search_order.value,
              certificate.value, block_size, table_size]
    )
 
    return "REACHABLE true" in stats, stats, cert



from tcheckerpy.tools import tck_compare, tck_liveness, tck_reach, tck_simulate, tck_syntax
import os

# test systems and oracle
test_systems_path = os.path.join(os.path.dirname(__file__), "examples")
with open(os.path.join(test_systems_path, "ad94.tck")) as file:
    system = file.read()
with open(os.path.join(test_systems_path, "ad94_witness.gv")) as file:
    witness = file.read()
with open(os.path.join(test_systems_path, "ad94_first_step.json")) as file:
    first_step = file.read()
with open(os.path.join(test_systems_path, "ad94.gv")) as file:
    dot_format = file.read()
with open(os.path.join(test_systems_path, "ad94.json")) as file:
    json_format = file.read()
with open(os.path.join(test_systems_path, "ad94_product.tck")) as file:
    product = file.read()

def test_tck_compare():
    result = tck_compare.compare(system, system, generate_witness = True)
    assert result[0]
    assert result[2] == witness

def test_tck_liveness():
    result = tck_liveness.liveness(system, tck_liveness.Algorithm.COUVSCC)
    assert not result[0]
    assert result[2] == ""

def test_tck_reach():
    assert not tck_reach.reach(system, tck_reach.Algorithm.REACH)[0]

def test_tck_simulate():
    assert tck_simulate.one_step_simulation(system) == first_step

def test_tck_syntax():
    tck_syntax.check(system)
    assert tck_syntax.to_dot(system) == dot_format
    assert tck_syntax.to_json(system) == json_format
    assert tck_syntax.create_product(system) == product